/*
 * Sweep, a sound wave editor.
 *
 * Copyright (C) 2000 Conrad Parker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <gtk/gtk.h>

#include <sndfile.h>

#include <sweep/sweep_types.h>
#include <sweep/sweep_sample.h>
#include <sweep/sweep_i18n.h>

#include "sweep_app.h"
#include "file_sndfile.h"
#include "sample.h"
#include "sample-display.h"

/* This source code can compile against both version 0 and version 1
** of libsndfile. The code for version 0 is at the top and the code
** for version 1 is below.
*/

extern int sndfile_save_options_dialog (sw_sample * sample);

static gchar* load_current_file = NULL; /* last visited dir for loading */
static gchar* save_current_file = NULL; /* last visited dir for saving */

typedef struct {
  sw_file_format_t file_format;
  gchar * name;
  gchar * exts; /* comma separated extensions */
} sw_file_format_desc;
    
typedef struct _sw_ff_instance sw_ff_instance;

struct _sw_ff_instance {
  sw_sample * sample;
  sw_file_format_t file_format;
};

static GtkWidget *
create_save_menu (sw_sample * sample) ;

static sw_ff_instance*
sw_ff_instance_new (sw_sample * sample, sw_file_format_t file_format) ;


static void
set_ffi_cb (GtkWidget * widget, gpointer data) ;

#if ! (defined (SNDFILE_1))

static sw_file_format_desc file_formats [] = {
  {
    SWEEP_FILE_FORMAT_RAW,
    "Raw PCM (headerless)",
    "raw",
  },
  {
    SWEEP_FILE_FORMAT_AIFF,
    "Apple/SGI AIFF,AIFF-C",
    "aiff,aifc"
  },
  {
    SWEEP_FILE_FORMAT_AU,
    "Sun/NeXT AU",
    "au",
  },
  {
    SWEEP_FILE_FORMAT_SVX,
    "Amiga IFF,SVX8,SVX16",
    "svx",
  },
  {
    SWEEP_FILE_FORMAT_PAF,
    "Ensoniq PARIS",
    "paf",
  },
  {
    SWEEP_FILE_FORMAT_IRCAM,
    "IRCAM/Berkeley/CARL SF",
    "sf",
  },
  {
    SWEEP_FILE_FORMAT_VOC,
    "Creative Labs VOC",
    "voc",
  },
  {
    SWEEP_FILE_FORMAT_WAV,
    "Microsoft WAV",
    "wav,riff",
  },
};

sw_file_format_t
guess_file_format (gchar * pathname)
{
  int i;
  sw_file_format_desc * desc;
  sw_file_format_t ret = SWEEP_FILE_FORMAT_WAV;
  gchar ** exts, **e;
  gchar * ext;

  ext = strrchr (pathname, '.');

  if (ext != NULL) {
    ext++;

    for (i = 0; i <  sizeof(file_formats)/sizeof(sw_file_format_desc); i++) {
      desc = &file_formats[i];
      exts = g_strsplit (desc->exts, ",", 16);
      for (e = exts; *e; e++) {
	if (!g_strncasecmp (*e, ext, SW_DIR_LEN)) {
	  ret = desc->file_format;
	}
      }
      g_strfreev (exts);
    }
  }

  return ret;
}    

static GtkWidget *
create_save_menu (sw_sample * sample)
{
  GtkWidget * menu;
  GtkWidget * menuitem;
  int i;
  sw_file_format_desc * desc;
  sw_ff_instance * ffi;

  menu = gtk_menu_new ();

  sample->file_format = SWEEP_FILE_FORMAT_BY_EXTENSION;

  menuitem = gtk_menu_item_new_with_label (_("By extension"));
  gtk_menu_append (GTK_MENU(menu), menuitem);
  ffi = sw_ff_instance_new (sample, SWEEP_FILE_FORMAT_BY_EXTENSION);
  gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
		      GTK_SIGNAL_FUNC(set_ffi_cb), ffi);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new(); /* Separator */
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_widget_show(menuitem);

  for (i = 0; i < sizeof(file_formats)/sizeof(sw_file_format_desc); i++) {
    desc = &file_formats[i];
    menuitem = gtk_menu_item_new_with_label (desc->name);
    gtk_menu_append (GTK_MENU(menu), menuitem);
    ffi = sw_ff_instance_new (sample, desc->file_format);
    gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
			GTK_SIGNAL_FUNC(set_ffi_cb), ffi);
    gtk_widget_show (menuitem);
  }

  return menu;
}

#endif 

sw_sample *
sample_load (char * pathname)
{
  g_free(load_current_file);
  load_current_file = g_strdup(pathname);

  return sndfile_sample_load (pathname);
}

static void
sample_load_ok_cb(GtkWidget * widget, gpointer data)
{
  gchar *dir;

  dir = gtk_file_selection_get_filename(GTK_FILE_SELECTION(data));

  sample_load(dir);

  gtk_widget_destroy(GTK_WIDGET(data));
}

static void
sample_load_cancel_cb(GtkWidget * widget, gpointer data)
{
  gtk_widget_destroy(GTK_WIDGET(data));
}

void
sample_load_cb(GtkWidget * wiget, gpointer data)
{
  GtkWidget *filesel;

  filesel = gtk_file_selection_new("Load Sample");
  gtk_window_set_position (GTK_WINDOW (filesel), GTK_WIN_POS_MOUSE);

  if(load_current_file) {
      gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), 
				      load_current_file);
  }

  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
		     "clicked", GTK_SIGNAL_FUNC(sample_load_ok_cb), filesel);

  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
		"clicked", GTK_SIGNAL_FUNC(sample_load_cancel_cb), filesel);

  gtk_widget_show(filesel);
}

void
sample_revert_cb (GtkWidget * widget, gpointer data)
{
  sw_view * view = (sw_view *)data;
  sw_sample * sample, * orig_sample;
  char pathname [SW_DIR_LEN];

  sample = view->sample;
  snprintf (pathname, SW_DIR_LEN,
	    "%s/%s", sample->directory,
	    sample->filename);

#ifdef DEBUG
  g_print ("Attempting to revert >>%s<<\n", pathname);
#endif

  orig_sample = sample_load (pathname);

#ifdef DEBUG
  g_print ("Loaded into %p\n", orig_sample);
#endif

  sample_replace_throughout (sample, orig_sample);

  sample_refresh_views (orig_sample);
}

static void
sample_save_as_ok_cb(GtkWidget * widget, gpointer data)
{
  sw_view * view = (sw_view *)data;
  sw_sample * sample;
  GtkWidget * filesel;
  gchar * pathname, * dn, * fn;

  sample = view->sample;
  filesel = gtk_widget_get_toplevel (widget);

  pathname = gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
  g_free(save_current_file);
  save_current_file = g_strdup(pathname);

  fn = g_basename (pathname);
  dn = g_dirname (pathname);

  sample_set_pathname (sample, dn, fn);

  sndfile_save_options_dialog (sample);

  gtk_widget_destroy(GTK_WIDGET(filesel));
}

static void
sample_save_as_cancel_cb(GtkWidget * widget, gpointer data)
{
  gtk_widget_destroy(GTK_WIDGET(data));
}

static sw_ff_instance *
sw_ff_instance_new (sw_sample * sample, sw_file_format_t file_format)
{
  sw_ff_instance * ffi;

  /* XXX: where to clean this up? */
  ffi = g_malloc (sizeof (sw_ff_instance));
  ffi->sample = sample;
  ffi->file_format = file_format;

  return ffi;
}

static void
set_ffi_cb (GtkWidget * widget, gpointer data)
{
  sw_ff_instance * ffi = (sw_ff_instance *)data;

  sample_set_file_format (ffi->sample, ffi->file_format);
}


void
sample_save_as_cb(GtkWidget * widget, gpointer data)
{
  sw_view * view = (sw_view *)data;
  sw_sample * sample;
  GtkWidget * filesel;
  GtkWidget * save_options;
  GtkWidget * frame;
  GtkWidget * hbox;
  GtkWidget * label;
  GtkWidget * option_menu;
  GtkWidget * save_menu;

  sample = view->sample;

  chdir(sample->directory);

  filesel = gtk_file_selection_new("Save Sample");
  gtk_window_set_position (GTK_WINDOW (filesel), GTK_WIN_POS_MOUSE);

  if(save_current_file)
      gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), 
				      save_current_file);
  save_options = gtk_hbox_new (TRUE, 1);

  frame = gtk_frame_new (_("Save Options"));
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
  gtk_box_pack_start (GTK_BOX (save_options), frame, TRUE, TRUE, 4);
  
  hbox = gtk_hbox_new (FALSE, 4);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 4);
  gtk_container_add (GTK_CONTAINER (frame), hbox);
  gtk_widget_show (hbox);
  
  label = gtk_label_new (_("Determine File Type:"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);
  
  option_menu = gtk_option_menu_new ();
  gtk_box_pack_start (GTK_BOX (hbox), option_menu, TRUE, TRUE, 0);
  gtk_widget_show (option_menu);
  
  save_menu = create_save_menu (sample);
  gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), save_menu);
  
  gtk_widget_show (frame);

  /* pack the containing save_options hbox into the save-dialog */
  gtk_box_pack_end (GTK_BOX (GTK_FILE_SELECTION (filesel)->main_vbox),
		    save_options, FALSE, FALSE, 0);

  gtk_widget_show (save_options);

  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
		     "clicked", GTK_SIGNAL_FUNC(sample_save_as_ok_cb), data);

  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
		"clicked", GTK_SIGNAL_FUNC(sample_save_as_cancel_cb), filesel);

  gtk_widget_show(filesel);
}

void
sample_save_cb(GtkWidget * widget, gpointer data)
{
  sw_view * view = (sw_view *)data;
  sw_sample * sample;

  sample = view->sample;

  if (!sample->filename) {
    sample_save_as_cb (widget, data);
  } else {
    sndfile_sample_save (sample);
  }
}

#ifdef SNDFILE_1
/*===========================================================================================
** libsndfile-1.0.X support below. 
*/

#include <sndfile.h>

sw_file_format_t
guess_file_format (gchar * pathname)
{
  int i;
  sw_file_format_t ret = SWEEP_FILE_FORMAT_WAV;
  gchar * ext;

  SF_FORMAT_INFO  info ;
  int	count ;

  ext = strrchr (pathname, '.');

  if (ext != NULL) {
    ext++;

    sf_command (NULL, SFC_GET_FORMAT_MAJOR_COUNT, &count, sizeof (int)) ;

    for (i = 0; i <  count ; i++) {
      info.format = i ;
      sf_command (NULL, SFC_GET_FORMAT_MAJOR, &info, sizeof (info)) ;
		
      if (!g_strncasecmp (ext, info.extension, SW_DIR_LEN)) {
	ret = info.format;
      }
    }
  }

  return ret;
}    

static GtkWidget *
create_save_menu (sw_sample * sample)
{
  GtkWidget * menu;
  GtkWidget * menuitem;
  int i;

  SF_FORMAT_INFO  info ;
  int	count ;

  sw_ff_instance * ffi;

  menu = gtk_menu_new ();

  sample->file_format = SWEEP_FILE_FORMAT_BY_EXTENSION;

  menuitem = gtk_menu_item_new_with_label (_("By extension"));
  gtk_menu_append (GTK_MENU(menu), menuitem);
  ffi = sw_ff_instance_new (sample, SWEEP_FILE_FORMAT_BY_EXTENSION);
  gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
		      GTK_SIGNAL_FUNC(set_ffi_cb), ffi);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new(); /* Separator */
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_widget_show(menuitem);

  sf_command (NULL, SFC_GET_FORMAT_MAJOR_COUNT, &count, sizeof (int)) ;

  for (i = 0; i < count ; i++) {
    info.format = i ;
    sf_command (NULL, SFC_GET_FORMAT_MAJOR, &info, sizeof (info)) ;
    
    menuitem = gtk_menu_item_new_with_label (info.name);
    gtk_menu_append (GTK_MENU(menu), menuitem);
    ffi = sw_ff_instance_new (sample, info.format);
    gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
			GTK_SIGNAL_FUNC(set_ffi_cb), ffi);
    gtk_widget_show (menuitem);
  }

  return menu;
}


#endif
